;;;; -*- Mode: Lisp;  -*-
;;;; *************************************************************************
;;;; Copyright (C) 2011 Roger Grantham
;;;; 
;;;; All rights reserved.
;;;; 
;;;; This file is part of Grammaticus.
;;;; 
;;;; Grammaticus is free software: you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by 
;;;; the Free Software Foundation, either version 3 of the License, or    
;;;; (at your option) any later version.                                  
;;;; 
;;;; Grammaticus is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with Grammaticus.  If not, see <http://www.gnu.org/licenses/>.
;;;; 
;;;; *************************************************************************
;;;;
;;;; Grammaticus Tokenizer
;;;;
;;;; The tokenizer is responsible for reading in text from an input
;;;; source and creating an input stream from which tokenized symbols
;;;; representing whole words and punctuation may be read.
;;;;
;;;; $Id:$
;;;;
;;;; *************************************************************************

(in-package :grammaticus-tokenizer)

(defparameter *alphabet* (list #\A #\Á #\Â #\B #\C #\D #\E #\É #\Ê #\F
			       #\G #\H #\I #\Í #\Î #\J #\K #\L #\M #\N 
			       #\O #\Ó #\Ô #\P #\Q #\R #\S #\T #\U #\Ú 
			       #\Û #\V #\X #\Y #\Ý #\Z #\a #\á #\â #\b 
			       #\c #\d #\e #\é #\ê #\f #\g #\h #\i #\í 
			       #\î #\j #\k #\l #\m #\n #\o #\ó #\ô #\p 
			       #\q #\r #\s #\t #\u #\ú #\û #\v #\x #\y 
			       #\ý #\z #\Æ #\æ ))

(defparameter *punctuation* (list #\. #\! #\? #\, #\; #\- #\( #\) #\" #\'))


(defun alphabetic-p (char)
  "Indicates whether the character belongs to the Latin alphabet."
  (if (null char)
      nil
      (member char *alphabet* :test #'char=)))

(defun punctuation-p (char)
  "Indicates whether the character is punctuation."
  (if (null char)
      nil
      (member char *punctuation* :test #'char=)))

(defun consumable-p (char)
  "Indicates whether the tokenizer should consume or skip the
character."
  (or (alphabetic-p char) (punctuation-p char)))


(defun make-token-stream (input-stream)  
  "Reads from an input stream and returns a token input stream from
   which the tokenized text (whole words and punctuation tokenized as
   symbols) may be read."
  (let ((word (read-raw-word input-stream)))
    (if word
	(cons (intern word)
	      #'(lambda ()
		  (let ((word (read-raw-word input-stream)))
		    (if (null word) 
			nil
			(intern word)))))
	nil)))



(defun read-raw-word (input-stream)
  "Reads from the input stream and collects the first
word. Non-alphabetic characters are discarded. Returns the raw word or
nil on EOF."
  ;; advance input stream to first alphabetic character or punctuation
  (do ((next-char (peek-char t input-stream nil nil) (peek-char t input-stream nil nil)))
      ((or (null next-char)(consumable-p next-char)))
    (unless (or (alphabetic-p next-char) (punctuation-p next-char))
      (read-char input-stream nil nil)))
  ;; read the first raw word
  (let ((raw-word (with-output-to-string (word)
		    (do ((char (read-char input-stream nil nil) 
				(read-char input-stream nil nil)))
			((not (consumable-p char)))
		      (unless (null char)
			(write-char char word))
		      (when (or (punctuation-p char) 
				(punctuation-p (peek-char nil input-stream nil nil)))
			(return))))))
    (if (string= "" raw-word) nil raw-word)))


(defun read-token (token-input-stream)
  "Reads the next token from the token input stream created by
make-token-stream. Returns nil at the end of the stream."
  (let ((token (car token-input-stream))
	(next-token (funcall (cdr token-input-stream))))
    (setf (car token-input-stream) next-token)
    token))
